SuperStrict

Rem
bbdoc: Holzchopf/Calc
End Rem
Module Holzchopf.Calc

ModuleInfo "Version: 1.00"
ModuleInfo "Author: Holzchopf"
ModuleInfo "License: Public Domain"

Import BRL.Blitz

Rem
bbdoc: Main calculator type
about: Contains some basic calculating functions.
End Rem
Type TCalc
	' ASCII codes
	Const NUMBERS_LB:Byte =		48
	Const NUMBERS_UB:Byte =		57
	Const POINT:Byte =			46
	Const CHARS_LB:Byte =		97
	Const CHARS_UB:Byte =		122
	Const CAPS_LB:Byte =		65
	Const CAPS_UB:Byte =		90
	Const COMMA:Byte =			44
	Const BRACKET_O:Byte =		40
	Const BRACKET_C:Byte =		41
	Const PLUS:Byte =			43
	Const MINUS:Byte =			45
	Const ADD:Byte =			43
	Const SUBTRACT:Byte =		45
	Const MULTIPLY:Byte =		42
	Const DIVIDE:Byte =			47
	Const EXPONENTIATE:Byte =	94

Rem
bbdoc: Array of saved #TCalcTerm terms.
End Rem
	Global gVarTerm:TCalcTerm[]
	Global _gArgVal:Double[]
	Global _gParName:String[]
	Global _gParseIndex:Int
	Global _gParseTerm:String
	Global _gMaxIndex:Int
	
	' allow overriding of named terms
Rem
bbdoc: True, if overriding of named terms is allowed.
End Rem
	Global gAllowOverride:Byte = False
Rem
bbdoc: Allows (true or no argument passed) or disables (false) overriding of named terms.<br>
Default is false.
End Rem
	Function AllowOverride(pAllow:Byte=True)
		gAllowOverride = pAllow
	End Function
	
	' use values for further calculations
	Global _gValuesID:Int
Rem
bbdoc: True, if term calls return their last computed value.
End Rem
	Global gUseValues:Byte = False
Rem
bbdoc: Forces terms to return their last computed value when called (true or no argument passed)
or to compute their value on every call (false).<br>
Default is false.
about: If UseValues is set, terms will be computed only once (when they're called the first time
after UseValues) optimizing calculating times in most cases.<br>
This feature is disabled by default because it's necessary to define when terms shall be
updated again (remember in loops!).<br>
Take a look at the sample <b>term generated image</b> to see how #UseValues works.
End Rem
	Function UseValues(pUse:Byte=True)
		gUseValues = pUse
		If pUse
			_gValuesID :+ 1
		EndIf
	End Function

	' ready-to-use parser
Rem
bbdoc: Parses a mathematical term.
returns: Result as double value.
End Rem
	Function Parse:Double(pTerm:String)
		_gMaxIndex = -1
		Local tmpExpr:TCalcExpression = ParseTerm(pTerm)
		If tmpExpr
			Return tmpExpr.Get()
		Else
			Return 0
		EndIf
	End Function

	' parse term
Rem
bbdoc: Parses a mathematical term.<br>
The term must be passed as string variable!
returns: Result as chain of #TCalcExpression objects.
End Rem
	Function ParseTerm:TCalcExpression(pTerm:String Var, pResetIndex:Byte=True, pPrec:Byte=0, pBracket:Byte=False, pComma:Byte=False)
		' start term index
		If pResetIndex
			_gParseIndex = 0
			_gParseTerm = pTerm
		EndIf
		Local expr:TCalcExpression
		Local i:Int
		Repeat
			Local c:Byte
			c = _GetNextChar(pTerm)
			' no valid value
			If c=0
				' mark: EXP_EXPRESSION @ 000
				TCalcException._Create(TCalcException.EXP_EXPRESSION, 000)
				Return Null
			'-----
			' sign
			'-----
			' negative value
			ElseIf c=MINUS And expr=Null
				pTerm = pTerm[1..]; _gParseIndex:+1
				Return TCalcSub.Create( TCalcValue.Create(0), ParseTerm(pTerm, False, , pBracket, pComma) )
			' explicit positive value
			ElseIf c=PLUS And expr=Null
				pTerm = pTerm[1..]; _gParseIndex:+1
				Return ParseTerm(pTerm, False, , pBracket, pComma)
			'-----------
			' expression
			'-----------
			' value
			ElseIf c>=NUMBERS_LB And c<=NUMBERS_UB And expr=Null
				' extract
				For i = 1 Until pTerm.length
					c = pTerm[i]
					If Not( (c>=NUMBERS_LB And c<=NUMBERS_UB) Or c=POINT )
						Exit
					EndIf
				Next
				expr = TCalcValue.Create( Double(pTerm[0..i]) )
				pTerm = pTerm[i..]; _gParseIndex:+i
			' named term
			ElseIf (c>=CHARS_LB And c<=CHARS_UB) Or (c>=CAPS_LB And c<=CAPS_UB) And expr=Null
				Local name:String = _GetName(pTerm)
				' search parameter
				expr = _FindParameter(name)
				' no parameter found
				If expr=Null
					Local term:TCalcTerm = FindNamedTerm(name)
					If term=Null
						' mark: INV_NAME @       001
						TCalcException._Create(TCalcException.INV_NAME, 001, _gParseIndex-name.length, _gParseIndex)
						Return Null
					EndIf
					' term without parameters
					If term.params.length=0
						expr = TCalcTermLink.Create(term)
					' term with parameters
					Else
						Local pars:TCalcExpression[] = _ParseParameters(pTerm, term.params.length)
						If pars=Null Return Null
						expr = TCalcTermLink.Create(term, pars)
					EndIf
				EndIf
			' bracket open
			ElseIf c=BRACKET_O And expr=Null
				pTerm = pTerm[1..]; _gParseIndex:+1
				expr = ParseTerm(pTerm, False, , True, pComma)
				If pTerm.length=0
					' mark: EXP_BRACKET @    002
					TCalcException._Create(TCalcException.EXP_BRACKET, 002)
					Return Null
				EndIf
				pTerm = pTerm[1..]; _gParseIndex:+1
			ElseIf expr=Null
				' mark: INV_CHAR @       003
				TCalcException._Create(TCalcException.INV_CHAR, 003)
				Return Null
			EndIf
			
			'-----------------------
			' term ending conditions
			c = _GetNextChar(pTerm)
			' term-inated ^_
			If c=0
				If pBracket
					' mark: EXP_BRACKET @    004
					TCalcException._Create(TCalcException.EXP_BRACKET, 004)
					Return Null
				ElseIf pComma
					' mark: EXP_COMMA @      005
					TCalcException._Create(TCalcException.EXP_COMMA, 005)
					Return Null
				EndIf
				Return expr
			' bracket close
			ElseIf c=BRACKET_C
				If pBracket
					Return expr
				Else
					' mark: INV_BRACKET @    006
					TCalcException._Create(TCalcException.INV_BRACKET, 006)
					Return Null
				EndIf
			' comma
			ElseIf c=COMMA
				If pComma
					Return expr
				Else
					' mark: INV_COMMA @      007
					TCalcException._Create(TCalcException.INV_COMMA, 007)
					Return Null
				EndIf
			' operation of lower order
			ElseIf _Precedence(c) <= pPrec
				Return expr
			EndIf
			
			'-----------
			' operators
			' addition
			If c=ADD
				pTerm = pTerm[1..]; _gParseIndex:+1
				Local expr2:TCalcExpression = ParseTerm(pTerm, False, 1, pBracket, pComma)
				If expr2=Null Return Null
				expr = TCalcAdd.Create( expr, expr2 )
			' subtraction
			ElseIf c=SUBTRACT
				pTerm = pTerm[1..]; _gParseIndex:+1
				Local expr2:TCalcExpression = ParseTerm(pTerm, False, 1, pBracket, pComma)
				If expr2=Null Return Null
				expr = TCalcSub.Create( expr, expr2 )
			' multiplication
			ElseIf c=MULTIPLY
				pTerm = pTerm[1..]; _gParseIndex:+1
				Local expr2:TCalcExpression = ParseTerm(pTerm, False, 2, pBracket, pComma)
				If expr2=Null Return Null
				expr = TCalcMul.Create( expr, expr2 )
			' division
			ElseIf c=DIVIDE
				pTerm = pTerm[1..]; _gParseIndex:+1
				Local expr2:TCalcExpression = ParseTerm(pTerm, False, 2, pBracket, pComma)
				If expr2=Null Return Null
				expr = TCalcDiv.Create( expr, expr2 )
			' exponentiation
			ElseIf c=EXPONENTIATE
				pTerm = pTerm[1..]; _gParseIndex:+1
				Local expr2:TCalcExpression = ParseTerm(pTerm, False, 3, pBracket, pComma)
				If expr2=Null Return Null
				expr = TCalcPow.Create( expr, expr2 )
			' multiplication la 5(3+3)
			ElseIf c=BRACKET_O
				pTerm = pTerm[1..]; _gParseIndex:+1
				Local expr2:TCalcExpression = ParseTerm(pTerm, False, , True, pComma)
				If expr2=Null Return Null
				expr = TCalcMul.Create( expr, expr2 )
				If pTerm.length=0
					' mark: EXP_BRACKET @    008
					TCalcException._Create(TCalcException.EXP_BRACKET, 008)
					Return Null
				EndIf
				pTerm = pTerm[1..]; _gParseIndex:+1
			' multiplication la 3sin(x)
			Else
				Local expr2:TCalcExpression = ParseTerm(pTerm, False, 2, pBracket, pComma)
				If expr2=Null Return Null
				expr = TCalcMul.Create( expr, expr2 )
			EndIf
		Until _GetNextChar(pTerm)=0
		Return expr
	End Function

	' create named link to variable
Rem
bbdoc: Creates a named link to a double variable.
about: Remember that names for links are always case-sensitive!
returns: #TCalcLinkVar
End Rem
	Function LinkVar:TCalcLinkVar(pVar:Double Var, pName:String)
		Local term:TCalcTerm = AllocTerm(pName)
		If term=Null Return Null
		Return term.LinkToVar(pVar)
	End Function
	' create named link to variable (int)
Rem
bbdoc: Creates a named link to an int variable.
about: Remember that names for links are always case-sensitive!
returns: #TCalcLinkVarInt
End Rem
	Function LinkVarInt:TCalcLinkVarInt(pVar:Int Var, pName:String)
		Local term:TCalcTerm = AllocTerm(pName)
		If term=Null Return Null
		Return term.LinkToVarInt(pVar)
	End Function
	' create named link to function
Rem
bbdoc: Creates a named link to a function which returns a double value.
about: Remember that names for links are always case-sensitive!
returns: #TCalcLinkFunc
End Rem
	Function LinkFunc:TCalcLinkFunc(pFunc:Double(), pName:String)
		Local term:TCalcTerm = AllocTerm(pName)
		If term=Null Return Null
		Return term.LinkToFunc(pFunc)
	End Function
	' create named link to function (double)
Rem
bbdoc: Creates a named link to a function which expects one argument (double) and returns a double value.
about: Remember that names for links are always case-sensitive!
returns: #TCalcLinkFuncD
End Rem
	Function LinkFuncD:TCalcLinkFuncD(pFunc:Double(a:Double), pName:String)
		Local term:TCalcTerm = AllocTerm(pName)
		If term=Null Return Null
		Return term.LinkToFuncD(pFunc)
	End Function
	' create named link to function (double, double)
Rem
bbdoc: Creates a named link to a function which expects two arguments (double) returns a double value.
about: Remember that names for links are always case-sensitive!
returns: #TCalcLinkFuncDD
End Rem
	Function LinkFuncDD:TCalcLinkFuncDD(pFunc:Double(a:Double, b:Double), pName:String)
		Local term:TCalcTerm = AllocTerm(pName)
		If term=Null Return Null
		Return term.LinkToFuncDD(pFunc)
	End Function
	' create named link to function (double, double, double)
Rem
bbdoc: Creates a named link to a function which expects three arguments (double) returns a double value.
about: Remember that names for links are always case-sensitive!
returns: #TCalcLinkFuncDDD
End Rem
	Function LinkFuncDDD:TCalcLinkFuncDDD(pFunc:Double(a:Double, b:Double, c:Double), pName:String)
		Local term:TCalcTerm = AllocTerm(pName)
		If term=Null Return Null
		Return term.LinkToFuncDDD(pFunc)
	End Function

	' allocate named term
Rem
bbdoc: Allocates space for a named term.
about: Remember that names for links are always case-sensitive!
returns: #TCalcTerm
End Rem
	Function AllocTerm:TCalcTerm(pName:String)
		Local term:TCalcTerm = New TCalcTerm
		Local i:Int = 0
		Local save:Byte = True
		' reset parse index
		_gParseIndex = 0
		_gParseTerm = pName
		' parse name and variables
		Repeat
			' get valid name
			Local name:String = _GetName(pName)
			If name=Null And term.name=Null
				' mark: NAME_EXPECTED @  020
				TCalcException._Create(TCalcException.NAME_EXPECTED, 020)
				Return Null
			ElseIf name=Null
				' mark: NAME_VARIABLE @  021
				TCalcException._Create(TCalcException.NAME_VARIABLE, 021)
				Return Null
			EndIf
			' extract name
			If term.name = Null
				' lookup term name
				Local trm:TCalcTerm = FindNamedTerm(name)
				If trm<>Null
					If gAllowOverride=False
						' mark: NAME_USED @      022
						TCalcException._Create(TCalcException.NAME_USED, 022, _gParseIndex-name.length, _gParseIndex)
						Return Null
					Else
						term = trm
						save = False
					EndIf
				EndIf
				term.name = name
			' extract variable
			Else
				term.params = term.params[..i+1]
				term.params[i] = name
				i:+1
			EndIf
			' trim white spaces
			Local c:Byte = _GetNextChar(pName)
			' no arguments at all
			If c=0 And i=0
				Exit
			' first: expect bracket open
			ElseIf c=BRACKET_O And i=0
				pName = pName[1..]; _gParseIndex:+1
				If _GetNextChar(pName)=0
					' mark: NAME_VARIABLE @  023
					TCalcException._Create(TCalcException.NAME_VARIABLE, 023)
					Return Null
				EndIf
			' after: expect comma
			ElseIf c=COMMA And i>0
				pName = pName[1..]; _gParseIndex:+1
				If _GetNextChar(pName)=0
					' mark: NAME_VARIABLE @  024
					TCalcException._Create(TCalcException.NAME_VARIABLE, 024)
					Return Null
				EndIf
			' last: expect bracket close
			ElseIf c=BRACKET_C
				pName = pName[1..]; _gParseIndex:+1
				If _GetNextChar(pName) <> 0
					' mark: INV_CHAR @       025
					TCalcException._Create(TCalcException.INV_CHAR, 025)
					Return Null
				EndIf
				Exit
			ElseIf c=0
				' mark: INV_END @        026
				TCalcException._Create(TCalcException.INV_END, 026)
				Return Null
			Else
				' mark: INV_CHAR @       027
				TCalcException._Create(TCalcException.INV_CHAR, 027)
				Return Null
			EndIf
		Until _GetNextChar(pName)=0
		' save term
		If save
			i = gVarTerm.length
			gVarTerm = gVarTerm[..i+1]
			gVarTerm[i] = term
			term._index = i
		EndIf
		Return term
	End Function
	
	' tidy up
Rem
bbdoc: Removes all named terms.
End Rem
	Function Clear()
		For Local i:Int = 0 Until gVarTerm.length
			gVarTerm[i].expr = Null
			gVarTerm[i] = Null
		Next
		gVarTerm = Null
	End Function
	
	' get precedence (order) of operation
	Function _Precedence:Byte( pOperation:Byte )
		Select pOperation
		Case ADD, SUBTRACT
			Return 1
		Case MULTIPLY, DIVIDE
			Return 2
		Case EXPONENTIATE
			Return 3
		Default
			Return 2
		End Select
	End Function
	
	' get parameters
	Function _ParseParameters:TCalcExpression[](pTerm:String Var, pCount:Int)
		Local s:Int = _gParseIndex
		Local i:Int = 0
		Local params:TCalcExpression[]
		Repeat
			Local c:Byte = _GetNextChar(pTerm)
			' begin (bracket open) or continue (comma) extraction
			If (i=0 And c=BRACKET_O) Or (c=COMMA And i>0)
				pTerm = pTerm[1..]; _gParseIndex:+1
				params = params[..i+1]
				params[i] = ParseTerm(pTerm, False, , True, True)
				If params[i]=Null Return Null
				i:+1
			' end (bracket close) extraction
			ElseIf c=BRACKET_C
				pTerm = pTerm[1..]; _gParseIndex:+1
				Exit
			ElseIf c=0
				' mark: ARG_EXPECTED @   040
				TCalcException._Create(TCalcException.ARG_EXPECTED, 040, s,s)
				Return Null
			Else
				' mark: ARG_BAD @        041
				TCalcException._Create(TCalcException.ARG_BAD, 041)
				Return Null
			EndIf
		Forever
		If i=pCount
			Return params
		ElseIf i>pCount
			' mark: ARG_MANY @       042
			TCalcException._Create(TCalcException.ARG_MANY, 042, s+1, _gParseIndex-1)
			Return Null
		ElseIf i<pCount
			' mark: ARG_FEW @        043
			TCalcException._Create(TCalcException.ARG_FEW, 043, s+1, _gParseIndex-1)
			Return Null
		EndIf
		Return params
	End Function
	
	' get next functional char of term
	' returns 0, if none was found
	Function _GetNextChar:Byte(pTerm:String Var)
		While pTerm.length>0
			' filter all white spaces
			If pTerm[0]>32
				Return pTerm[0]
			EndIf
			pTerm = pTerm[1..]; _gParseIndex:+1
		Wend
		Return 0
	End Function
	
	' get valid name
	' autotrims leading white spaces
	Function _GetName:String(pTerm:String Var)
		Local i:Int
		' remove leading white spaces
		_GetNextChar(pTerm)
		For i = 0 Until pTerm.length
			Local c:Byte = pTerm[i]
			' letters
			 ' following numbers
			 ' underline
			If Not((c>=CHARS_LB And c<=CHARS_UB) Or (c>=CAPS_LB And c<=CAPS_UB)..
			 Or ((c>=NUMBERS_LB And c<=NUMBERS_UB) And i>0)..
			 Or c=95)
				Exit
			EndIf
		Next
		Local name:String = pTerm[..i]
		pTerm = pTerm[i..]; _gParseIndex:+i
		Return name
	End Function
	
	' look up parameters
	Function _FindParameter:TCalcExpression(pName:String)
		For Local i:Int = 0 Until _gParName.length
			If _gParName[i] = pName
				Return TCalcExprParameter.Create(i)
			EndIf
		Next
	End Function
	
	' look up named term
Rem
bbdoc: Finds term by name.
about: Remember that names for links are always case-sensitive!
returns: #TCalcTerm
End Rem
	Function FindNamedTerm:TCalcTerm(pName:String)
		For Local i:Int = 0 Until gVarTerm.length
			If i = _gMaxIndex Return Null
			If gVarTerm[i].name = pName
				Return gVarTerm[i]
			EndIf
		Next
	End Function
End Type

Rem
bbdoc: Error handling type.
End Rem
Type TCalcException
	' general term errors
	Const EXP_EXPRESSION:Byte =	$10
	Const INV_NAME:Byte =		$11
	Const EXP_BRACKET:Byte =		$12
	Const EXP_COMMA:Byte =		$13
	Const INV_CHAR:Byte =		$14
	Const INV_BRACKET:Byte =		$15
	Const INV_COMMA:Byte =		$16
	Const INV_END:Byte =		$17
	' argument errors
	Const ARG_EXPECTED:Byte =	$20
	Const ARG_BAD:Byte =		$21
	Const ARG_MANY:Byte =		$22
	Const ARG_FEW:Byte =		$23
	' naming errors
	Const NAME_EXPECTED:Byte =	$40
	Const NAME_USED:Byte =		$41
	Const NAME_VARIABLE:Byte =	$42
	Const NAME_MANY:Byte =		$43
	Const NAME_FEW:Byte =		$44
	' call errors
	Const CALL_MANY:Byte =		$80
	Const CALL_FEW:Byte =		$81
	Const CALL_UNDEF:Byte =		$82
	Const CALL_INV:Byte =		$83

Rem
bbdoc: Error code of last occured error.
about: The codes are defined as const in #TCalcException.
End Rem
	Field code:Byte
Rem
bbdoc: Reference to the point, where the error was recognized.<br>
Mainly used for debugging.
End Rem
	Field ref:Byte
Rem
bbdoc: Start index in term string of incorrect chars.
End Rem
	Field indexS:Int
Rem
bbdoc: End index in term string of incorrect chars.
End Rem
	Field indexE:Int
Rem
bbdoc: Term that caused the error.
End Rem
	Field term:String
	
	' behaviour
Rem
bbdoc: True, if exceptions are thrown on errors.
End Rem
	Global gThrow:Byte = True
Rem
bbdoc: #TCalcException object of last occured error (only when ThrowOnError is not set).
End Rem
	Global gError:TCalcException
Rem
bbdoc: Defines, whether errors throw an exception (true or no argument passed) or not (false).<br>
Default is true.
about: When #ThrowOnError is not set, errors in terms will cause an 0 value result but no exception
will be thrown. You have to check for errors manually by checking #gError.
End Rem
	Function ThrowOnError(pThrow:Byte=True)
		gThrow = pThrow
	End Function
	
	' throws an complete calc exception
	Function _Create(pCode:Byte, pRef:Byte, pStart:Int=-1, pEnd:Int=-1)
		Local exc:TCalcException = New TCalcException
		exc.code = pCode
		exc.ref = pRef
		If pStart=-1
			exc.indexS = TCalc._gParseIndex
		Else
			exc.indexS = pStart
		EndIf
		If pEnd=-1
			exc.indexE = TCalc._gParseIndex+1
		Else
			exc.indexE = pEnd
		EndIf
		exc.term = TCalc._gParseTerm
		If gThrow
			Throw exc
		Else
			gError = exc
		EndIf
	End Function
	
	' nice built-in ToString()
Rem
bbdoc: Nice ToString() method.
returns: Error description.
End Rem
	Method ToString:String()
		Local message:String
		DebugStop
		Local codeStr:String = "00" +code
		codeStr = codeStr[codeStr.length-3..]
		Local refStr:String = "00" +ref
		refStr = refStr[refStr.length-3..]
		
		Select code
		Case EXP_EXPRESSION
			message = "Expecting expression"
		Case INV_NAME
			message = "Named expression not found"
		Case EXP_BRACKET
			message = "Expecting ')'"
		Case EXP_COMMA
			message = "Expecting ','"
		Case INV_CHAR
			message = "Unexpected char occured"
		Case INV_BRACKET
			message = "Unexpected ')' occured"
		Case INV_COMMA
			message = "Unexpected ',' occured"
		Case INV_END
			message = "Unexpected end of term"
		Case ARG_EXPECTED
			message = "Argument expected"
		Case ARG_BAD
			message = "Bad argument"
		Case ARG_MANY
			message = "Too many arguments"
		Case ARG_FEW
			message = "Too few arguments"
		Case NAME_EXPECTED
			message = "Name expected"
		Case NAME_USED
			message = "Name already in use"
		Case NAME_VARIABLE
			message = "Variable expected"
		Case NAME_MANY
			message = "Too many parameters"
		Case NAME_FEW
			message = "Too few parameters"
		Case CALL_MANY
			message = "called with too many arguments"
		Case CALL_FEW
			message = "called with too few arguments"
		Case CALL_UNDEF
			message = "contains undefined named expression"
		Case CALL_INV
			message = "Invalid function call"
		End Select
		If code<$80
			Return "#"+codeStr+"-"+refStr+"~n"+message+" in '"+term+"'~n"+term[..indexS]+" >"+term[indexS..indexE]+"< "+term[indexE..]+"~n"
		Else
			Return "#"+codeStr+"-"+refStr+"~n'"+term+"' "+message+"~n"
		EndIf
	End Method
End Type

' calc interface
Rem
bbdoc: Type for terms.
about: In Holzchopf.Calc every named thing is considered to be a term (variables, functions or
other terms). So basically you can do anything just by using #TCalcTerm objects.
End Rem
Type TCalcTerm
	Field name:String
	Field params:String[]
	Field expr:TCalcExpression
	Field term:String
	Field _index:Int
	Field _value:Double, _valueID:Int
	
	' allocate named term
Rem
bbdoc: Allocates space for a named term.
about: Remember that names for links are always case-sensitive!
returns: #TCalcTerm
End Rem
	Function Alloc:TCalcTerm(pName:String)
		Return TCalc.AllocTerm(pName)
	End Function
	' remove term from index
Rem
bbdoc: Removes an term from the list of named terms.
about: Only necessary when #AllowOverride is disabled.
End Rem
	Method Free()
		For Local i:Int = 0 Until TCalc.gVarTerm.length
			If TCalc.gVarTerm[i] = Self
				TCalc.gVarTerm = TCalc.gVarTerm[..i] +TCalc.gVarTerm[i+1..]
				Exit
			EndIf
		Next
	End Method
	
	' create a new term
Rem
bbdoc: Creates a new term (can be named or unnamed).
about: Remember that names for links are always case-sensitive!
returns: #TCalcTerm
End Rem
	Function Create:TCalcTerm(pTerm:String, pName:String=Null)
		Local term:TCalcTerm
		' unnamed expression
		If pName=Null
			term = New TCalcTerm
			term._index = -1
		' named expression
		Else
			term = Alloc(pName)
		EndIf
		' set expression
		term.Set(pTerm)
		Return term
	End Function

	' set expression
Rem
bbdoc: Sets a new string.
End Rem
	Method Set(pTerm:String)
		term = pTerm
		Parse()
	End Method
Rem
bbdoc: Parses the set string.
about: Usually done automatically. Use this function to update specific terms when #UseValues is enabled.
End Rem
	Method Parse()
		Local tempTerm:String = term
		TCalc._gMaxIndex = _index
		' put parameters on global stack and parse expression
		TCalc._gParName = params
		expr = TCalc.ParseTerm(tempTerm)
		TCalc._gParName = Null
		TCalc._gMaxIndex = -1
	End Method
	
	' link to var
Rem
bbdoc: Links the term to a double variable.
returns: #TCalcLinkVar
End Rem
	Method LinkToVar:TCalcLinkVar(pVar:Double Var)
		expr = TCalcLinkVar.Create(pVar)
		Return TCalcLinkVar(expr)
	End Method
	' link to var (int)
Rem
bbdoc: Links the term to an int variable.
returns: #TCalcLinkVarInt
End Rem
	Method LinkToVarInt:TCalcLinkVarInt(pVar:Int Var)
		expr = TCalcLinkVarInt.Create(pVar)
		Return TCalcLinkVarInt(expr)
	End Method
	' link to function
Rem
bbdoc: Links the term to a double returning function with no arguments.
returns: #TCalcLinkFunc
End Rem
	Method LinkToFunc:TCalcLinkFunc(pFunc:Double())
		If params.length>0
			' mark: NAME_MANY @      100
			TCalcException._Create(TCalcException.NAME_MANY, 100, 0, TCalc._gParseIndex)
			Return Null
		EndIf
		expr = TCalcLinkFunc.Create(pFunc)
		Return TCalcLinkFunc(expr)
	End Method
	' link to function (double)
Rem
bbdoc: Links the term to a double returning function with one argument (double).
returns: #TCalcLinkFuncD
End Rem
	Method LinkToFuncD:TCalcLinkFuncD(pFunc:Double(a:Double))
		If params.length<1
			' mark: NAME_FEW @       110
			TCalcException._Create(TCalcException.NAME_FEW, 110, 0, TCalc._gParseIndex)
			Return Null
		ElseIf params.length>1
			' mark: NAME_MANY @      111
			TCalcException._Create(TCalcException.NAME_MANY, 111, 0, TCalc._gParseIndex)
			Return Null
		EndIf
		expr = TCalcLinkFuncD.Create(pFunc)
		Return TCalcLinkFuncD(expr)
	End Method
	' link to function (double, double)
Rem
bbdoc: Links the term to a double returning function with two arguments (double).
returns: #TCalcLinkFuncDD
End Rem
	Method LinkToFuncDD:TCalcLinkFuncDD(pFunc:Double(a:Double, b:Double))
		If params.length<2
			' mark: NAME_FEW @       120
			TCalcException._Create(TCalcException.NAME_FEW, 120, 0, TCalc._gParseIndex)
			Return Null
		ElseIf params.length>2
			' mark: NAME_MANY @      121
			TCalcException._Create(TCalcException.NAME_MANY, 121, 0, TCalc._gParseIndex)
			Return Null
		EndIf
		expr = TCalcLinkFuncDD.Create(pFunc)
		Return TCalcLinkFuncDD(expr)
	End Method
	' link to function (double, double, double)
Rem
bbdoc: Links the term to a double returning function with three arguments (double).
returns: #TCalcLinkFuncDDD
End Rem
	Method LinkToFuncDDD:TCalcLinkFuncDDD(pFunc:Double(a:Double, b:Double, c:Double))
		If params.length<3
			' mark: NAME_FEW @       130
			TCalcException._Create(TCalcException.NAME_FEW, 130, 0, TCalc._gParseIndex)
			Return Null
		ElseIf params.length>3
			' mark: NAME_MANY @      131
			TCalcException._Create(TCalcException.NAME_MANY, 131, 0, TCalc._gParseIndex)
			Return Null
		EndIf
		expr = TCalcLinkFuncDDD.Create(pFunc)
		Return TCalcLinkFuncDDD(expr)
	End Method
	
	' get expression value
Rem
bbdoc: Returns the result of this term.
about: When getting a term with arguments, those must be passed as array of type double.
returns: Double
End Rem
	Method Get:Double(pParams:Double[]=Null)
		If expr=Null
			' mark: CALL_INV @       190
			TCalcException._Create(TCalcException.CALL_INV, 190)
			Return 0
		EndIf
		If pParams.length<params.length
			' mark: CALL_FEW @       191
			TCalcException._Create(TCalcException.CALL_FEW, 191)
			Return 0
		ElseIf pParams.length>params.length
			' mark: CALL_MANY @      192
			TCalcException._Create(TCalcException.CALL_MANY, 192)
			Return 0
		EndIf
		' push on stack
		Local tempArgs:Double[] = TCalc._gArgVal
		TCalc._gArgVal = pParams
		' calc expression
		Local result:Double = expr.Get()
		' pop params
		TCalc._gArgVal = tempArgs
		' return result
		Return result
	End Method
End Type

' base expression
Rem
bbdoc: Base expression type.
about: Every object in a term is considered to be an expression (can be a number, a variable,
another term like 1+2 or a link to a function parameter).
End Rem
Type TCalcExpression Abstract
Rem
returns: Double
End Rem
	Method Get:Double() Abstract
End Type
			
' parameter
Type TCalcExprParameter Extends TCalcExpression
	Field index:Int
	Function Create:TCalcExprParameter(pIndex:Int)
		Local par:TCalcExprParameter = New TCalcExprParameter
		par.index = pIndex
		Return par
	End Function
	Method Get:Double()
		Return TCalc._gArgVal[index]
	End Method
End Type

' link to expression
Type TCalcTermLink Extends TCalcExpression
	Field term:TCalcTerm
	Field params:TCalcExpression[]
	Function Create:TCalcTermLink(pTerm:TCalcTerm, pParams:TCalcExpression[]=Null)
		Local lnk:TCalcTermLink = New TCalcTermLink
		lnk.term = pTerm
		lnk.params = pParams
		Return lnk
	End Function
	Method Get:Double()
		' use old value if possible
		If TCalc.gUseValues=True And TCalc._gValuesID = term._valueID
			Return term._value
		EndIf
		term._valueID = TCalc._gValuesID
		If term.expr=Null
			' mark: CALL_UNDEF @     200
			TCalcException._Create(TCalcException.CALL_UNDEF, 2000, 0,0)
			Return 0
		EndIf
		' calc params
		Local args:Double[]
		For Local i:Int = 0 Until params.length
			args = args[..i+1]
			args[i] = params[i].Get()
		Next
		' push on stack
		Local tempArgs:Double[] = TCalc._gArgVal
		TCalc._gArgVal = args
		' calc expression
		Local result:Double = term.expr.Get()
		term._value = result
		' pop params
		TCalc._gArgVal = tempArgs
		' return result
		Return result
	End Method
End Type
' link to variable
Type TCalcLinkVar Extends TCalcExpression
	Field variable:Double Ptr
	Function Create:TCalcLinkVar(pVar:Double Var)
		Local lv:TCalcLinkVar = New TCalcLinkVar
		lv.variable = Varptr pVar
		Return lv
	End Function
	Method Get:Double()
		Return variable[0]
	End Method
	Method Relink(pVar:Double Var)
		variable = Varptr pVar
	End Method
End Type
' link to variable (int)
Type TCalcLinkVarInt Extends TCalcExpression
	Field variable:Int Ptr
	Function Create:TCalcLinkVarInt(pVar:Int Var)
		Local lv:TCalcLinkVarInt = New TCalcLinkVarInt
		lv.variable = Varptr pVar
		Return lv
	End Function
	Method Get:Double()
		Return Double(variable[0])
	End Method
	Method Relink(pVar:Int Var)
		variable = Varptr pVar
	End Method
End Type
' link to function ()
Type TCalcLinkFunc Extends TCalcExpression
	Field func:Double()
	Function Create:TCalcLinkFunc(pFunc:Double())
		Local fn:TCalcLinkFunc = New TCalcLinkFunc
		fn.func = pFunc
		Return fn
	End Function
	Method Get:Double()
		Return func()
	End Method
	Method Relink(pFunc:Double())
		func = pFunc
	End Method
End Type
' link to function (double)
Type TCalcLinkFuncD Extends TCalcExpression
	Field func:Double(a:Double)
	Function Create:TCalcLinkFuncD(pFunc:Double(a:Double))
		Local fn:TCalcLinkFuncD = New TCalcLinkFuncD
		fn.func = pFunc
		Return fn
	End Function
	Method Get:Double()
		If TCalc._gArgVal.length=0
			' mark: CALL_INV @       210
			TCalcException._Create(TCalcException.CALL_INV, 210)
			Return 0
		EndIf
		Return func(TCalc._gArgVal[0])
	End Method
	Method Relink(pFunc:Double(a:Double))
		func = pFunc
	End Method
End Type
' link to function (double, double)
Type TCalcLinkFuncDD Extends TCalcExpression
	Field func:Double(a:Double,b:Double)
	Function Create:TCalcLinkFuncDD(pFunc:Double(a:Double,b:Double))
		Local fn:TCalcLinkFuncDD = New TCalcLinkFuncDD
		fn.func = pFunc
		Return fn
	End Function
	Method Get:Double()
		If TCalc._gArgVal.length<2
			' mark: CALL_INV @       220
			TCalcException._Create(TCalcException.CALL_INV, 220)
			Return 0
		EndIf
		Return func(TCalc._gArgVal[0],TCalc._gArgVal[1])
	End Method
	Method Relink(pFunc:Double(a:Double,b:Double))
		func = pFunc
	End Method
End Type
' link to function (double, double, double)
Type TCalcLinkFuncDDD Extends TCalcExpression
	Field func:Double(a:Double,b:Double,c:Double)
	Function Create:TCalcLinkFuncDDD(pFunc:Double(a:Double,b:Double,c:Double))
		Local fn:TCalcLinkFuncDDD = New TCalcLinkFuncDDD
		fn.func = pFunc
		Return fn
	End Function
	Method Get:Double()
		If TCalc._gArgVal.length<3
			' mark: CALL_INV @       230
			TCalcException._Create(TCalcException.CALL_INV, 230)
			Return 0
		EndIf
		Return func(TCalc._gArgVal[0],TCalc._gArgVal[1],TCalc._gArgVal[2])
	End Method
	Method Relink(pFunc:Double(a:Double,b:Double,c:Double))
		func = pFunc
	End Method
End Type

' value
Type TCalcValue Extends TCalcExpression
	Field value:Double
	Function Create:TCalcValue(pValue:Double)
		Local val:TCalcValue = New TCalcValue
		val.value = pValue
		Return val
	End Function
	Method Get:Double()
		Return value
	End Method
End Type

' addition
Type TCalcAdd Extends TCalcExpression
	Field a:TCalcExpression
	Field b:TCalcExpression
	Function Create:TCalcAdd(pA:TCalcExpression, pB:TCalcExpression)
		Local add:TCalcAdd = New TCalcAdd
		add.a = pA
		add.b = pB
		Return add
	End Function
	Method Get:Double()
		Return a.Get() + b.Get()
	End Method
End Type
' subtraction
Type TCalcSub Extends TCalcExpression
	Field a:TCalcExpression
	Field b:TCalcExpression
	Function Create:TCalcSub(pA:TCalcExpression, pB:TCalcExpression)
		Local sub:TCalcSub = New TCalcSub
		sub.a = pA
		sub.b = pB
		Return sub
	End Function
	Method Get:Double()
		Return a.Get() - b.Get()
	End Method
End Type
' multiplication
Type TCalcMul Extends TCalcExpression
	Field a:TCalcExpression
	Field b:TCalcExpression
	Function Create:TCalcMul(pA:TCalcExpression, pB:TCalcExpression)
		Local mul:TCalcMul = New TCalcMul
		mul.a = pA
		mul.b = pB
		Return mul
	End Function
	Method Get:Double()
		Return a.Get() * b.Get()
	End Method
End Type
' divistion
Type TCalcDiv Extends TCalcExpression
	Field a:TCalcExpression
	Field b:TCalcExpression
	Function Create:TCalcDiv(pA:TCalcExpression, pB:TCalcExpression)
		Local div:TCalcDiv = New TCalcDiv
		div.a = pA
		div.b = pB
		Return div
	End Function
	Method Get:Double()
		Return a.Get() / b.Get()
	End Method
End Type
' exponentiation
Type TCalcPow Extends TCalcExpression
	Field a:TCalcExpression
	Field b:TCalcExpression
	Function Create:TCalcPow(pA:TCalcExpression, pB:TCalcExpression)
		Local pow:TCalcPow = New TCalcPow
		pow.a = pA
		pow.b = pB
		Return pow
	End Function
	Method Get:Double()
		Return a.Get() ^ b.Get()
	End Method
End Type

